import { z } from "zod";
import bcrypt from "bcrypt";

import {
  createTRPCRouter,
  protectedProcedure,
  publicProcedure,
} from "~/server/api/trpc";
import { TRPCError } from "@trpc/server";

import { UserModel } from "~/models/UserModel";

// Main login router for backend
export const loginRouter = createTRPCRouter({
    // Takes in username and password
  create_account: publicProcedure
    .input(z.object({email: z.string(), username: z.string(), password: z.string()}))
    .mutation(async ({ ctx, input }) => {
      // Encryption using bcrypt salt and pepper hashing
      const user = await UserModel.findOne({username: input.username});
      const userEmail = await UserModel.findOne({email: input.email});
      if (user) {
        throw new TRPCError ({
          code: "BAD_REQUEST",
          message: "Username is already taken!",
        })
      }
      if (userEmail) {
        throw new TRPCError ({
          code: "BAD_REQUEST",
          message: "Email is already taken!",
        })
      }
      const saltRounds = 10;
      const hashedPassword: string = await bcrypt.hash(input.password, saltRounds);
      // Store in database
      const newUser = await UserModel.create({username: input.username,
        email: input.email ,
        password: hashedPassword });
      await newUser.save();
      // Return greeting Żnformation
        return {
        user_info: newUser
      };
    }),

  Login: publicProcedure
    .input(z.object({ username: z.string(), password: z.string()}))
    .mutation(async ({ ctx, input }) => {
      // Encryption using bcrypt salt and pepper hashing
      const saltRounds = 10;
      // Get hash from database
      const user = await UserModel.findOne({username: input.username});
      if (!user) {
        throw new TRPCError ({
          code: "BAD_REQUEST",
          message: "User doesn't exist",
        })
      }
      const isPasswordValid = await bcrypt.compare(input.password, user.password);
      if (!isPasswordValid) {
        throw new TRPCError ({
          code: "BAD_REQUEST",
          message: "Username or Password is incorrect",
        })
      }
      // Return for sign in
      return {
        user: user,
        success: true,
        message: "Successfully logged in!",
      }
    }),
});
