import { z } from "zod";
import bcrypt from "bcrypt";
import { type Action } from '@finvis/usxi'

import {
  createTRPCRouter,
  protectedProcedure,
  publicProcedure,
} from "~/server/api/trpc";
import { UserModel } from "~/models/UserModel";
import { TRPCError } from "@trpc/server";


// Main login router for backend
export const userRouter = createTRPCRouter({

    // Function to add a Api Key to a user
    addUserApiKey: protectedProcedure
    .input(z.object({username: z.string(), apiKey: z.object({name: z.string(), key: z.string(), iv: z.string()}) }))
    .mutation(async ({ ctx, input }) => {
      const user = await UserModel.findOne({username: input.username});

      // Check if user exists
      if (!user) {
        throw new TRPCError ({
          code: "BAD_REQUEST",
          message: "User not found",
        })
      }

      // Add the new API key to the user's array
      user.apiKeys.push(input.apiKey);

      // Save new user
      await user.save();

      // Return greeting information
      return {
        user_info: user
      };
    }),

    // Function to get a users Api Keys
    getUserApiKeys: protectedProcedure
    .input(z.object({ username: z.string()}))
    .query(async ({ ctx, input }) => {
      const user = await UserModel.findOne({username: input.username});

      // Check if user exists
      if (!user) {
        throw new TRPCError ({
          code: "BAD_REQUEST",
          message: "User doesn't exist",
        })
      }

      // Check if user has API keys
      if (!user.apiKeys || user.apiKeys.length) {
        throw new TRPCError ({
            code: "BAD_REQUEST",
            message: "User has no api keys",
        })
      }

      // Return a list of the users API keys
      return user.apiKeys
    }),

    // Function to change a users password
    changeUserPass: protectedProcedure
    .input(z.object({username: z.string(), oldPass: z.string(), newPass: z.string(), newConfirm: z.string()}))
    .mutation(async ({ ctx, input }) => {
      const user = await UserModel.findOne({username: input.username});

      // Check if user exists
      if (!user) {
        throw new TRPCError ({
          code: "BAD_REQUEST",
          message: "User not found",
        })
      }

      // Check password validity
      const isPasswordValid = await bcrypt.compare(input.oldPass, user.password);
      if (!isPasswordValid) {
        throw new TRPCError ({
          code: "BAD_REQUEST",
          message: "Username or Password is incorrect",
        })
      }

      // Check new passwords
      if (input.newPass !== input.newConfirm) {
        throw new TRPCError ({
          code: "BAD_REQUEST",
          message: "New passwords do not match",
        })
      }

      const saltRounds = 10;
      const hashedPassword: string = await bcrypt.hash(input.newPass, saltRounds);

      // Update password
      user.password = hashedPassword

      // Save new user information
      await user.save();

      // Return updated information
      return {
        user_info: user
      };
    }),

    // Add new balance to user
    putUserBalanceHist: protectedProcedure
    .input(z.object({username: z.string(), balance: z.object({ amount: z.number(), date: z.string() })}))
    .mutation(async ({ ctx, input }) => {
      const user = await UserModel.findOne({username: input.username});

      // Check if user exists
      if (!user) {
        throw new TRPCError ({
          code: "BAD_REQUEST",
          message: "User not found",
        })
      }

      user.balance.push(input.balance);

      // Save new user information
      await user.save();

      // Return updated information
      return {
        user_info: user
      };
    }),

    // Function to get a users balance history
    getUserBalanceHist: protectedProcedure
    .input(z.object({ username: z.string(), amount: z.number() }))
    .query(async ({ ctx, input }) => {
      const user = await UserModel.findOne({username: input.username});

      // Check if user exists
      if (!user) {
        throw new TRPCError ({
          code: "BAD_REQUEST",
          message: "User doesn't exist",
        })
      }

      // Check if user has API keys
      if (!user.balance || user.balance.length === 0) {
        throw new TRPCError ({
            code: "BAD_REQUEST",
            message: "User has no balance history",
        })
      }

      if(user.balance.length < input.amount) {
        throw new TRPCError ({
            code: "BAD_REQUEST",
            message: "User only has length ${user.balance.length} balance history",
        })
      }

      if(input.amount === -1){
        return user.balance
      }

      const lastBalances = await UserModel.findOne({username: input.username})
      // Select only balance
      .select('balance')
      // Get last n items
      .slice('balance', -input.amount)
      // Convert to JavaScript object
      .lean()

      return lastBalances
    }),


    // Add new trade to user
    putUserTradeHist: protectedProcedure
    .input(z.object({username: z.string(), trade: z.custom<Action>()}))
    .mutation(async ({ ctx, input }) => {
      const user = await UserModel.findOne({username: input.username});

      // Check if user exists
      if (!user) {
        throw new TRPCError ({
          code: "BAD_REQUEST",
          message: "User not found",
        })
      }

      user.trades.push(input.trade);

      // Save new user information
      await user.save();

      // Return updated information
      return {
        user_info: user
      };
    }),

    // Function to get a users Trade history
    getUserTradeHist: protectedProcedure
    .input(z.object({ username: z.string(), amount: z.number() }))
    .query(async ({ ctx, input }) => {
      const user = await UserModel.findOne({username: input.username});

      // Check if user exists
      if (!user) {
        throw new TRPCError ({
          code: "BAD_REQUEST",
          message: "User doesn't exist",
        })
      }

      // Check if user has API keys
      if (!user.trades || user.trades.length === 0) {
        throw new TRPCError ({
            code: "BAD_REQUEST",
            message: "User has no trades history",
        })
      }

      if(user.trades.length < input.amount) {
        throw new TRPCError ({
            code: "BAD_REQUEST",
            message: "User only has length ${user.trades.length} balance history",
        })
      }

      if(input.amount === -1){
        return user.trades
      }

      const lastTrades = await UserModel.findOne({username: input.username})
      // Select only balance
      .select('trades')
      // Get last n items
      .slice('trades', -input.amount)
      // Convert to JavaScript object
      .lean()

      return lastTrades
    }),


    // Add new position to user
    putUserPositions: protectedProcedure
    .input(z.object({username: z.string(), position: z.object({ symbol: z.string(),
      quantity: z.number(), marketValue: z.number(), costBasis: z.number(), pricePerShare: z.number(), })}))
    .mutation(async ({ ctx, input }) => {
      const user = await UserModel.findOne({username: input.username});

      // Check if user exists
      if (!user) {
        throw new TRPCError ({
          code: "BAD_REQUEST",
          message: "User not found",
        })
      }

      user.positions.push(input.position);

      // Save new user information
      await user.save();

      // Return updated information
      return {
        user_info: user
      };
    }),

    // Function to get a users positions
    getUserPositions: protectedProcedure
    .input(z.object({ username: z.string()}))
    .query(async ({ ctx, input }) => {
      const user = await UserModel.findOne({username: input.username});

      // Check if user exists
      if (!user) {
        throw new TRPCError ({
          code: "BAD_REQUEST",
          message: "User doesn't exist",
        })
      }

      // Check if user has API keys
      if (!user.positions || user.positions.length === 0) {
        throw new TRPCError ({
            code: "BAD_REQUEST",
            message: "User has no positions",
        })
      }

      return user.positions
    }),


    //Paper trading function


    // Add new balance to user
    putPaperBalanceHist: protectedProcedure
    .input(z.object({username: z.string(), balance: z.object({ amount: z.number(), date: z.string() })}))
    .mutation(async ({ ctx, input }) => {
      const user = await UserModel.findOne({username: input.username});

      // Check if user exists
      if (!user) {
        throw new TRPCError ({
          code: "BAD_REQUEST",
          message: "User not found",
        })
      }

      user.paperbalance.push(input.balance);

      // Save new user information
      await user.save();

      // Return updated information
      return {
        user_info: user
      };
    }),

    // Function to get a users balance history
    getPaperBalanceHist: protectedProcedure
    .input(z.object({ username: z.string(), amount: z.number() }))
    .query(async ({ ctx, input }) => {
      const user = await UserModel.findOne({username: input.username});

      // Check if user exists
      if (!user) {
        throw new TRPCError ({
          code: "BAD_REQUEST",
          message: "User doesn't exist",
        })
      }

      // Check if user has API keys
      if (!user.paperbalance || user.paperbalance.length === 0) {
        throw new TRPCError ({
            code: "BAD_REQUEST",
            message: "User has no balance history",
        })
      }

      if(user.paperbalance.length < input.amount) {
        throw new TRPCError ({
            code: "BAD_REQUEST",
            message: "User only has length ${user.balance.length} balance history",
        })
      }

      if(input.amount === -1){
        return user.paperbalance
      }

      const lastBalances = await UserModel.findOne({username: input.username})
      // Select only balance
      .select('paperbalance')
      // Get last n items
      .slice('paperbalance', -input.amount)
      // Convert to JavaScript object
      .lean()

      return lastBalances
    }),


    // Add new trade to user
    putPaperTradeHist: protectedProcedure
    .input(z.object({username: z.string(), trade: z.custom<Action>()}))
    .mutation(async ({ ctx, input }) => {
      const user = await UserModel.findOne({username: input.username});

      // Check if user exists
      if (!user) {
        throw new TRPCError ({
          code: "BAD_REQUEST",
          message: "User not found",
        })
      }

      user.papertrades.push(input.trade);

      // Save new user information
      await user.save();

      // Return updated information
      return {
        user_info: user
      };
    }),

    // Function to get a users Trade history
    getPaperTradeHist: protectedProcedure
    .input(z.object({ username: z.string(), amount: z.number() }))
    .query(async ({ ctx, input }) => {
      const user = await UserModel.findOne({username: input.username});

      // Check if user exists
      if (!user) {
        throw new TRPCError ({
          code: "BAD_REQUEST",
          message: "User doesn't exist",
        })
      }

      // Check if user has API keys
      if (!user.papertrades || user.papertrades.length === 0) {
        throw new TRPCError ({
            code: "BAD_REQUEST",
            message: "User has no trades history",
        })
      }

      if(user.papertrades.length < input.amount) {
        throw new TRPCError ({
            code: "BAD_REQUEST",
            message: "User only has length ${user.trades.length} balance history",
        })
      }

      if(input.amount === -1){
        return user.papertrades
      }

      const lastTrades = await UserModel.findOne({username: input.username})
       // Select only balance
      .select('papertrades')
       // Get last n items
      .slice('papertrades', -input.amount)
       // Convert to JavaScript object
      .lean()

      return lastTrades
    }),


    // Add new position to user
    putPaperPositions: protectedProcedure
    .input(z.object({username: z.string(), position: z.object({ symbol: z.string(),
      quantity: z.number(), marketValue: z.number(), costBasis: z.number(), pricePerShare: z.number(), })}))
    .mutation(async ({ ctx, input }) => {
      const user = await UserModel.findOne({username: input.username});

      // Check if user exists
      if (!user) {
        throw new TRPCError ({
          code: "BAD_REQUEST",
          message: "User not found",
        })
      }

      user.paperpositions.push(input.position);

      // Save new user information
      await user.save();

      // Return updated information
      return {
        user_info: user
      };
    }),

    // Function to get a users positions
    getPaperPositions: protectedProcedure
    .input(z.object({ username: z.string()}))
    .query(async ({ ctx, input }) => {
      const user = await UserModel.findOne({username: input.username});

      // Check if user exists
      if (!user) {
        throw new TRPCError ({
          code: "BAD_REQUEST",
          message: "User doesn't exist",
        })
      }

      // Check if user has positions
      if (!user.paperpositions || user.paperpositions.length === 0) {
        throw new TRPCError ({
            code: "BAD_REQUEST",
            message: "User has no positions",
        })
      }

      return user.paperpositions
    }),
});
